import NIO
import Logging

extension MQTTPacket {
    struct Connect: MQTTPacketOutboundType {
        private static let protocolName = "MQTT"
        private static let protocolLevel: UInt8 = 0x04 // 3.1.1
        
        private let configurationWrapper: ConfigurationWrapper
        
        init(configuration: MQTTConfiguration) {
            configurationWrapper = ConfigurationWrapper(configuration: configuration)
        }
        
        var configuration: MQTTConfiguration {
            return configurationWrapper.configuration
        }
        
        func serialize() throws -> MQTTPacket {
            var buffer = ByteBufferAllocator().buffer(capacity: 0)
            
            // Variable header
            try buffer.writeMQTTString(Self.protocolName, "Protocol name")
            buffer.writeInteger(Self.protocolLevel)
            
            // leave room for flags
            let flagsIndex = buffer.writerIndex
            buffer.moveWriterIndex(forwardBy: 1)
            
            if configuration.keepAliveInterval <= .seconds(0) {
                buffer.writeInteger(UInt16(0))
            } else if configuration.keepAliveInterval >= .seconds(Int64(UInt16.max)) {
                buffer.writeInteger(UInt16.max)
            } else {
                let seconds = configuration.keepAliveInterval.nanoseconds / TimeAmount.seconds(1).nanoseconds
                buffer.writeInteger(UInt16(seconds))
            }
            
            // Payload
            let flags = try serializePayload(into: &buffer)
            buffer.setInteger(flags.rawValue, at: flagsIndex)
            
            return MQTTPacket(kind: .connect, data: buffer)
        }
        
        private func serializePayload(into buffer: inout ByteBuffer) throws -> Flags {
            var flags: Flags = []
            
            if configuration.cleanSession {
                flags.insert(.cleanSession)
            }
            
            try buffer.writeMQTTString(configuration.clientId, "Client Identifier")
            
            if let lastWillMessage = configuration.lastWillMessage {
                flags.insert(.containsLastWill)
                try buffer.writeMQTTString(lastWillMessage.topic, "Topic")
                
                if var payload = lastWillMessage.payload {
                    try buffer.writeMQTTDataWithLength(&payload, "Last Will Payload")
                } else {
                    buffer.writeInteger(UInt16(0))
                }
                
                switch lastWillMessage.qos {
                case .atMostOnce:
                    break
                case .atLeastOnce:
                    flags.insert(.lastWillQoS1)
                case .exactlyOnce:
                    flags.insert(.lastWillQoS2)
                }
                
                if lastWillMessage.retain {
                    flags.insert(.lastWillRetain)
                }
            }
            
            if let credentials = configuration.credentials {
                flags.insert(.containsUsername)
                try buffer.writeMQTTString(credentials.username, "Username")
                
                if var password = credentials.password {
                    flags.insert(.containsPassword)
                    try buffer.writeMQTTDataWithLength(&password, "Password")
                }
            }
            
            return flags
        }
    }
}

extension MQTTPacket.Connect {
    // Wrapper to avoid heap allocations when added to NIOAny
    fileprivate class ConfigurationWrapper {
        let configuration: MQTTConfiguration
        
        init(configuration: MQTTConfiguration) {
            self.configuration = configuration
        }
    }
    
    fileprivate struct Flags: OptionSet {
        let rawValue: UInt8
        
        static let cleanSession = Flags(rawValue: 1 << 1)
        static let containsLastWill = Flags(rawValue: 1 << 2)
        static let lastWillQoS1 = Flags(rawValue: 1 << 3)
        static let lastWillQoS2 = Flags(rawValue: 1 << 4)
        static let lastWillRetain = Flags(rawValue: 1 << 5)
        static let containsPassword = Flags(rawValue: 1 << 6)
        static let containsUsername = Flags(rawValue: 1 << 7)
        
        init(rawValue: UInt8) {
            self.rawValue = rawValue
        }
    }
}
